@@ -1,23 +1,25 @@
-from __future__ import unicode_literals
-
-import hashlib
+import html
 import json
-import time
+import string
 import warnings
+from urllib.parse import quote, urlparse
 
-from django import forms
 from django.conf import settings
 from django.contrib import messages
 from django.contrib.auth import (
     authenticate,
     get_backends,
+    get_user_model,
     login as django_login,
     logout as django_logout,
 )
 from django.contrib.auth.models import AbstractUser
-from django.contrib.auth.password_validation import validate_password
+from django.contrib.auth.password_validation import (
+    MinimumLengthValidator,
+    validate_password,
+)
 from django.contrib.sites.shortcuts import get_current_site
-from django.core.cache import cache
+from django.core.exceptions import FieldDoesNotExist
 from django.core.mail import EmailMessage, EmailMultiAlternatives
 from django.http import HttpResponse, HttpResponseRedirect
 from django.shortcuts import resolve_url
@@ -25,37 +27,61 @@ from django.template import TemplateDoesNotExist
 from django.template.loader import render_to_string
 from django.urls import reverse
 from django.utils import timezone
+from django.utils.crypto import get_random_string
 from django.utils.encoding import force_str
 from django.utils.translation import gettext_lazy as _
 
-from ..utils import (
+from allauth import app_settings as allauth_app_settings
+from allauth.account import signals
+from allauth.account.app_settings import AuthenticationMethod
+from allauth.core import context, ratelimit
+from allauth.core.internal.adapter import BaseAdapter
+from allauth.utils import (
     build_absolute_uri,
-    email_address_exists,
     generate_unique_username,
-    get_user_model,
     import_attribute,
 )
+
 from . import app_settings
 
 
-class DefaultAccountAdapter(object):
+class DefaultAccountAdapter(BaseAdapter):
+    """The adapter class allows you to override various functionality of the
+    ``allauth.account`` app.  To do so, point ``settings.ACCOUNT_ADAPTER`` to
+    your own class that derives from ``DefaultAccountAdapter`` and override the
+    behavior by altering the implementation of the methods according to your own
+    needs.
+    """
 
     error_messages = {
+        "account_inactive": _("This account is currently inactive."),
+        "duplicate_email": _(
+            "This email address is already associated with this account."
+        ),
+        "email_password_mismatch": _(
+            "The email address and/or password you specified are not correct."
+        ),
+        "email_taken": _("A user is already registered with this email address."),
+        "enter_current_password": _("Please type your current password."),
+        "incorrect_code": _("Incorrect code."),
+        "incorrect_password": _("Incorrect password."),
+        "invalid_password_reset": _("The password reset token was invalid."),
+        "max_email_addresses": _("You cannot add more than %d email addresses."),
+        "too_many_login_attempts": _(
+            "Too many failed login attempts. Try again later."
+        ),
+        "unknown_email": _("The email address is not assigned to any user account"),
         "username_blacklisted": _(
             "Username can not be used. Please use other username."
         ),
+        "username_password_mismatch": _(
+            "The username and/or password you specified are not correct."
+        ),
         "username_taken": AbstractUser._meta.get_field("username").error_messages[
             "unique"
         ],
-        "too_many_login_attempts": _(
-            "Too many failed login attempts. Try again later."
-        ),
-        "email_taken": _("A user is already registered with this e-mail address."),
     }
 
-    def __init__(self, request=None):
-        self.request = request
-
     def stash_verified_email(self, request, email):
         request.session["account_verified_email"] = email
 
@@ -82,24 +108,53 @@ class DefaultAccountAdapter(object):
             ret = verified_email.lower() == email.lower()
         return ret
 
+    def can_delete_email(self, email_address):
+        from allauth.account.models import EmailAddress
+
+        has_other = (
+            EmailAddress.objects.filter(user_id=email_address.user_id)
+            .exclude(pk=email_address.pk)
+            .exists()
+        )
+        login_by_email = (
+            app_settings.AUTHENTICATION_METHOD == AuthenticationMethod.EMAIL
+        )
+        if email_address.primary:
+            if has_other:
+                # Don't allow, let the user mark one of the others as primary
+                # first.
+                return False
+            elif login_by_email:
+                # Last email & login is by email, prevent dangling account.
+                return False
+            return True
+        elif has_other:
+            # Account won't be dangling.
+            return True
+        elif login_by_email:
+            # This is the last email.
+            return False
+        else:
+            return True
+
     def format_email_subject(self, subject):
         prefix = app_settings.EMAIL_SUBJECT_PREFIX
         if prefix is None:
-            site = get_current_site(self.request)
+            site = get_current_site(context.request)
             prefix = "[{name}] ".format(name=site.name)
         return prefix + force_str(subject)
 
     def get_from_email(self):
         """
-        This is a hook that can be overridden to programatically
+        This is a hook that can be overridden to programmatically
         set the 'from' email address for sending emails
         """
         return settings.DEFAULT_FROM_EMAIL
 
-    def render_mail(self, template_prefix, email, context):
+    def render_mail(self, template_prefix, email, context, headers=None):
         """
-        Renders an e-mail to `email`.  `template_prefix` identifies the
-        e-mail that is to be sent, e.g. "account/email/email_confirmation"
+        Renders an email to `email`.  `template_prefix` identifies the
+        email that is to be sent, e.g. "account/email/email_confirmation"
         """
         to = [email] if isinstance(email, str) else email
         subject = render_to_string("{0}_subject.txt".format(template_prefix), context)
@@ -110,29 +165,39 @@ class DefaultAccountAdapter(object):
         from_email = self.get_from_email()
 
         bodies = {}
-        for ext in ["html", "txt"]:
+        html_ext = app_settings.TEMPLATE_EXTENSION
+        for ext in [html_ext, "txt"]:
             try:
                 template_name = "{0}_message.{1}".format(template_prefix, ext)
                 bodies[ext] = render_to_string(
                     template_name,
                     context,
-                    self.request,
+                    globals()["context"].request,
                 ).strip()
             except TemplateDoesNotExist:
                 if ext == "txt" and not bodies:
                     # We need at least one body
                     raise
         if "txt" in bodies:
-            msg = EmailMultiAlternatives(subject, bodies["txt"], from_email, to)
-            if "html" in bodies:
-                msg.attach_alternative(bodies["html"], "text/html")
+            msg = EmailMultiAlternatives(
+                subject, bodies["txt"], from_email, to, headers=headers
+            )
+            if html_ext in bodies:
+                msg.attach_alternative(bodies[html_ext], "text/html")
         else:
-            msg = EmailMessage(subject, bodies["html"], from_email, to)
+            msg = EmailMessage(
+                subject, bodies[html_ext], from_email, to, headers=headers
+            )
             msg.content_subtype = "html"  # Main content is now text/html
         return msg
 
     def send_mail(self, template_prefix, email, context):
-        msg = self.render_mail(template_prefix, email, context)
+        ctx = {
+            "email": email,
+            "current_site": get_current_site(globals()["context"].request),
+        }
+        ctx.update(context)
+        msg = self.render_mail(template_prefix, email, ctx)
         msg.send()
 
     def get_signup_redirect_url(self, request):
@@ -167,7 +232,7 @@ class DefaultAccountAdapter(object):
 
     def get_email_confirmation_redirect_url(self, request):
         """
-        The URL to return to after successful e-mail confirmation.
+        The URL to return to after successful email confirmation.
         """
         if request.user.is_authenticated:
             if app_settings.EMAIL_CONFIRMATION_AUTHENTICATED_REDIRECT_URL:
@@ -177,6 +242,14 @@ class DefaultAccountAdapter(object):
         else:
             return app_settings.EMAIL_CONFIRMATION_ANONYMOUS_REDIRECT_URL
 
+    def get_password_change_redirect_url(self, request):
+        """
+        The URL to redirect to after a successful password change/set.
+
+        NOTE: Not called during the password reset flow.
+        """
+        return reverse("account_change_password")
+
     def is_open_for_signup(self, request):
         """
         Checks whether or not the site is open for signups.
@@ -237,6 +310,8 @@ class DefaultAccountAdapter(object):
             user_field(user, "last_name", last_name)
         if "password1" in data:
             user.set_password(data["password1"])
+        elif "password" in data:
+            user.set_password(data["password"])
         else:
             user.set_unusable_password()
         self.populate_username(request, user)
@@ -259,27 +334,14 @@ class DefaultAccountAdapter(object):
             ub.lower() for ub in app_settings.USERNAME_BLACKLIST
         ]
         if username.lower() in username_blacklist_lower:
-            raise forms.ValidationError(self.error_messages["username_blacklisted"])
+            raise self.validation_error("username_blacklisted")
         # Skipping database lookups when shallow is True, needed for unique
         # username generation.
         if not shallow:
             from .utils import filter_users_by_username
 
             if filter_users_by_username(username).exists():
-                user_model = get_user_model()
-                username_field = app_settings.USER_MODEL_USERNAME_FIELD
-                error_message = user_model._meta.get_field(
-                    username_field
-                ).error_messages.get("unique")
-                if not error_message:
-                    error_message = self.error_messages["username_taken"]
-                raise forms.ValidationError(
-                    error_message,
-                    params={
-                        "model_name": user_model.__name__,
-                        "field_label": username_field,
-                    },
-                )
+                raise self.validation_error("username_taken")
         return username
 
     def clean_email(self, email):
@@ -295,40 +357,41 @@ class DefaultAccountAdapter(object):
         restric the allowed password choices.
         """
         min_length = app_settings.PASSWORD_MIN_LENGTH
-        if min_length and len(password) < min_length:
-            raise forms.ValidationError(
-                _("Password must be a minimum of {0} " "characters.").format(min_length)
-            )
+        if min_length:
+            MinimumLengthValidator(min_length).validate(password)
         validate_password(password, user)
         return password
 
     def validate_unique_email(self, email):
-        if email_address_exists(email):
-            raise forms.ValidationError(self.error_messages["email_taken"])
         return email
 
     def add_message(
         self,
         request,
         level,
-        message_template,
+        message_template=None,
         message_context=None,
         extra_tags="",
+        message=None,
     ):
         """
         Wrapper of `django.contrib.messages.add_message`, that reads
         the message text from a template.
         """
         if "django.contrib.messages" in settings.INSTALLED_APPS:
+            if message:
+                messages.add_message(request, level, message, extra_tags=extra_tags)
+                return
             try:
                 if message_context is None:
                     message_context = {}
-                message = render_to_string(
+                escaped_message = render_to_string(
                     message_template,
                     message_context,
-                    self.request,
+                    context.request,
                 ).strip()
-                if message:
+                if escaped_message:
+                    message = html.unescape(escaped_message)
                     messages.add_message(request, level, message, extra_tags=extra_tags)
             except TemplateDoesNotExist:
                 pass
@@ -380,6 +443,54 @@ class DefaultAccountAdapter(object):
             form_spec["field_order"].append(field.html_name)
         return form_spec
 
+    def pre_login(
+        self,
+        request,
+        user,
+        *,
+        email_verification,
+        signal_kwargs,
+        email,
+        signup,
+        redirect_url
+    ):
+        if not user.is_active:
+            return self.respond_user_inactive(request, user)
+
+    def post_login(
+        self,
+        request,
+        user,
+        *,
+        email_verification,
+        signal_kwargs,
+        email,
+        signup,
+        redirect_url
+    ):
+        from .utils import get_login_redirect_url
+
+        response = HttpResponseRedirect(
+            get_login_redirect_url(request, redirect_url, signup=signup)
+        )
+
+        if signal_kwargs is None:
+            signal_kwargs = {}
+        signals.user_logged_in.send(
+            sender=user.__class__,
+            request=request,
+            response=response,
+            user=user,
+            **signal_kwargs,
+        )
+        self.add_message(
+            request,
+            messages.SUCCESS,
+            "account/messages/logged_in.txt",
+            {"user": user},
+        )
+        return response
+
     def login(self, request, user):
         # HACK: This is not nice. The proper Django way is to use an
         # authentication backend
@@ -407,35 +518,57 @@ class DefaultAccountAdapter(object):
         """
         Marks the email address as confirmed on the db
         """
-        email_address.verified = True
-        email_address.set_as_primary(conditional=True)
-        email_address.save()
+        from allauth.account.internal.flows import manage_email
+
+        return manage_email.confirm_email(request, email_address)
 
     def set_password(self, user, password):
         user.set_password(password)
         user.save()
 
     def get_user_search_fields(self):
-        user = get_user_model()()
-        return filter(
-            lambda a: a and hasattr(user, a),
-            [
-                app_settings.USER_MODEL_USERNAME_FIELD,
-                "first_name",
-                "last_name",
-                "email",
-            ],
-        )
+        ret = []
+        User = get_user_model()
+        candidates = [
+            app_settings.USER_MODEL_USERNAME_FIELD,
+            "first_name",
+            "last_name",
+            "email",
+        ]
+        for candidate in candidates:
+            try:
+                User._meta.get_field(candidate)
+                ret.append(candidate)
+            except FieldDoesNotExist:
+                pass
+        return ret
 
     def is_safe_url(self, url):
-        try:
-            from django.utils.http import url_has_allowed_host_and_scheme
-        except ImportError:
-            from django.utils.http import (
-                is_safe_url as url_has_allowed_host_and_scheme,
-            )
+        from django.utils.http import url_has_allowed_host_and_scheme
 
-        return url_has_allowed_host_and_scheme(url, allowed_hosts=None)
+        # get_host already validates the given host, so no need to check it again
+        allowed_hosts = {context.request.get_host()} | set(settings.ALLOWED_HOSTS)
+
+        if "*" in allowed_hosts:
+            parsed_host = urlparse(url).netloc
+            allowed_host = {parsed_host} if parsed_host else None
+            return url_has_allowed_host_and_scheme(url, allowed_hosts=allowed_host)
+
+        return url_has_allowed_host_and_scheme(url, allowed_hosts=allowed_hosts)
+
+    def get_reset_password_from_key_url(self, key):
+        """
+        Method intented to be overriden in case the password reset email
+        needs to point to your frontend/SPA.
+        """
+        # We intentionally accept an opaque `key` on the interface here, and not
+        # implementation details such as a separate `uidb36` and `key. Ideally,
+        # this should have done on `urls` level as well.
+        path = reverse(
+            "account_reset_password_from_key", kwargs={"uidb36": "UID", "key": "KEY"}
+        )
+        path = path.replace("UID-KEY", quote(key))
+        return build_absolute_uri(self.request, path)
 
     def get_email_confirmation_url(self, request, emailconfirmation):
         """Constructs the email confirmation (activation) url.
@@ -448,13 +581,31 @@ class DefaultAccountAdapter(object):
         ret = build_absolute_uri(request, url)
         return ret
 
+    def should_send_confirmation_mail(self, request, email_address, signup):
+        send_email = ratelimit.consume(
+            request,
+            action="confirm_email",
+            key=email_address.email.lower(),
+        )
+        return send_email
+
+    def send_account_already_exists_mail(self, email):
+        signup_url = build_absolute_uri(context.request, reverse("account_signup"))
+        password_reset_url = build_absolute_uri(
+            context.request, reverse("account_reset_password")
+        )
+        ctx = {
+            "request": context.request,
+            "signup_url": signup_url,
+            "password_reset_url": password_reset_url,
+        }
+        self.send_mail("account/email/account_already_exists", email, ctx)
+
     def send_confirmation_mail(self, request, emailconfirmation, signup):
-        current_site = get_current_site(request)
         activate_url = self.get_email_confirmation_url(request, emailconfirmation)
         ctx = {
             "user": emailconfirmation.email_address.user,
             "activate_url": activate_url,
-            "current_site": current_site,
             "key": emailconfirmation.key,
         }
         if signup:
@@ -472,31 +623,20 @@ class DefaultAccountAdapter(object):
     def _get_login_attempts_cache_key(self, request, **credentials):
         site = get_current_site(request)
         login = credentials.get("email", credentials.get("username", "")).lower()
-        login_key = hashlib.sha256(login.encode("utf8")).hexdigest()
-        return "allauth/login_attempts@{site_id}:{login}".format(
-            site_id=site.pk, login=login_key
-        )
+        return "{site}:{login}".format(site=site.domain, login=login)
 
     def _delete_login_attempts_cached_email(self, request, **credentials):
-        if app_settings.LOGIN_ATTEMPTS_LIMIT:
-            cache_key = self._get_login_attempts_cache_key(request, **credentials)
-            cache.delete(cache_key)
+        cache_key = self._get_login_attempts_cache_key(request, **credentials)
+        ratelimit.clear(request, action="login_failed", key=cache_key)
 
     def pre_authenticate(self, request, **credentials):
-        if app_settings.LOGIN_ATTEMPTS_LIMIT:
-            cache_key = self._get_login_attempts_cache_key(request, **credentials)
-            login_data = cache.get(cache_key, None)
-            if login_data:
-                dt = timezone.now()
-                current_attempt_time = time.mktime(dt.timetuple())
-                if len(
-                    login_data
-                ) >= app_settings.LOGIN_ATTEMPTS_LIMIT and current_attempt_time < (
-                    login_data[-1] + app_settings.LOGIN_ATTEMPTS_TIMEOUT
-                ):
-                    raise forms.ValidationError(
-                        self.error_messages["too_many_login_attempts"]
-                    )
+        cache_key = self._get_login_attempts_cache_key(request, **credentials)
+        if not ratelimit.consume(
+            request,
+            action="login_failed",
+            key=cache_key,
+        ):
+            raise self.validation_error("too_many_login_attempts")
 
     def authenticate(self, request, **credentials):
         """Only authenticates, does not actually login. See `login`"""
@@ -507,19 +647,28 @@ class DefaultAccountAdapter(object):
         user = authenticate(request, **credentials)
         alt_user = AuthenticationBackend.unstash_authenticated_user()
         user = user or alt_user
-        if user and app_settings.LOGIN_ATTEMPTS_LIMIT:
+        if user:
             self._delete_login_attempts_cached_email(request, **credentials)
         else:
             self.authentication_failed(request, **credentials)
         return user
 
     def authentication_failed(self, request, **credentials):
-        if app_settings.LOGIN_ATTEMPTS_LIMIT:
-            cache_key = self._get_login_attempts_cache_key(request, **credentials)
-            data = cache.get(cache_key, [])
-            dt = timezone.now()
-            data.append(time.mktime(dt.timetuple()))
-            cache.set(cache_key, data, app_settings.LOGIN_ATTEMPTS_TIMEOUT)
+        pass
+
+    def reauthenticate(self, user, password):
+        from allauth.account.models import EmailAddress
+        from allauth.account.utils import user_username
+
+        credentials = {"password": password}
+        username = user_username(user)
+        if username:
+            credentials["username"] = username
+        email = EmailAddress.objects.get_primary_email(user)
+        if email:
+            credentials["email"] = email
+        reauth_user = self.authenticate(context.request, **credentials)
+        return reauth_user is not None and reauth_user.pk == user.pk
 
     def is_ajax(self, request):
         return any(
@@ -530,6 +679,79 @@ class DefaultAccountAdapter(object):
             ]
         )
 
+    def get_client_ip(self, request):
+        x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
+        if x_forwarded_for:
+            ip = x_forwarded_for.split(",")[0]
+        else:
+            ip = request.META.get("REMOTE_ADDR")
+        return ip
+
+    def get_http_user_agent(self, request):
+        return request.META.get("HTTP_USER_AGENT", "Unspecified")
+
+    def generate_emailconfirmation_key(self, email):
+        key = get_random_string(64).lower()
+        return key
+
+    def get_login_stages(self):
+        ret = []
+        ret.append("allauth.account.stages.EmailVerificationStage")
+        if allauth_app_settings.MFA_ENABLED:
+            ret.append("allauth.mfa.stages.AuthenticateStage")
+        return ret
+
+    def get_reauthentication_methods(self, user):
+        """The order of the methods returned matters. The first method is the
+        default when using the `@reauthentication_required` decorator.
+        """
+        ret = []
+        if not user.is_authenticated:
+            return ret
+        if user.has_usable_password():
+            ret.append(
+                {
+                    "description": _("Use your password"),
+                    "url": reverse("account_reauthenticate"),
+                }
+            )
+        if allauth_app_settings.MFA_ENABLED:
+            from allauth.mfa.utils import is_mfa_enabled
+
+            if is_mfa_enabled(user):
+                ret.append(
+                    {
+                        "description": _("Use your authenticator app"),
+                        "url": reverse("mfa_reauthenticate"),
+                    }
+                )
+        return ret
+
+    def send_notification_mail(self, template_prefix, user, context=None, email=None):
+        from allauth.account.models import EmailAddress
+
+        if not app_settings.EMAIL_NOTIFICATIONS:
+            return
+        if not email:
+            email = EmailAddress.objects.get_primary_email(user)
+        if not email:
+            return
+        ctx = {
+            "timestamp": timezone.now(),
+            "ip": self.get_client_ip(self.request),
+            "user_agent": self.get_http_user_agent(self.request),
+        }
+        if context:
+            ctx.update(context)
+        self.send_mail(template_prefix, email, ctx)
+
+    def generate_login_code(self):
+        forbidden_chars = "0OI18B2Z"
+        allowed_chars = string.ascii_uppercase + string.digits
+        for ch in forbidden_chars:
+            allowed_chars = allowed_chars.replace(ch, "")
+        return get_random_string(length=6, allowed_chars=allowed_chars)
+
 
 def get_adapter(request=None):
     return import_attribute(app_settings.ADAPTER)(request)
